%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%%%%% Elliptic solver for Poisson equation %%%%%
%%%%%     to calculate the nondivergent    %%%%%
%%%%%      barotropic stream function      %%%%%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%Note can be efficiently improved with the 
%Gauss Siedel version, pag 


%free slip condition
% free=1
%no slip
 free=0

%spheric Laplacian
  spheric=1
%no spheric normal dx and dy
%  spheric=0

%%%%%%%% Interpolation of depth on the PSI-Q-grid
 for ir=2:jdm-1
     for ic=2:idm-1
        depthq(ir,ic)=0.25*(depth(ir,ic)+depth(ir,ic-1)+depth(ir-1,ic)+depth(ir-1,ic-1));
     end
 end
 for ir=2:jdm-1
     depthq(ir,  1)=0.5*(depth(ir,    1)+depth(ir-1,    1));
     depthq(ir,idm)=0.5*(depth(ir,idm-1)+depth(ir-1,idm-1));
 end
 for ic=2:idm-1
     depthq(  1,ic)=0.5*(depth(    1,ic)+depth(    1,ic-1));
     depthq(jdm,ic)=0.5*(depth(jdm-1,ic)+depth(jdm-1,ic-1));
 end
 depthq(  1,  1)=depth(    1,    1);
 depthq(jdm,  1)=depth(jdm-1,    1);
 depthq(  1,idm)=depth(    1,idm-1);
 depthq(jdm,idm)=depth(jdm-1,idm-1);

%%%%%%%% Interpolation of shh on the PSI-Q-grid
 for ir=2:jdm-1
     for ic=2:idm-1
        sshq(ir,ic)=0.25*(srfhgt(ir,ic)+srfhgt(ir,ic-1)+srfhgt(ir-1,ic)+srfhgt(ir-1,ic-1));
     end
 end
 for ir=2:jdm-1
     sshq(ir,  1)=0.5*(srfhgt(ir,    1)+srfhgt(ir-1,    1));
     sshq(ir,idm)=0.5*(srfhgt(ir,idm-1)+srfhgt(ir-1,idm-1));
 end
 for ic=2:idm-1
     sshq(  1,ic)=0.5*(srfhgt(    1,ic)+srfhgt(    1,ic-1));
     sshq(jdm,ic)=0.5*(srfhgt(jdm-1,ic)+srfhgt(jdm-1,ic-1));
 end
 sshq(  1,  1)=srfhgt(    1,    1);
 sshq(jdm,  1)=srfhgt(jdm-1,    1);
 sshq(  1,idm)=srfhgt(    1,idm-1);
 sshq(jdm,idm)=srfhgt(jdm-1,idm-1);


%$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$
%calculation of the vorticity

%remove Nan on land and allow below interpolation
 v_btrop(  1,:)=0;
 v_btrop(jdm,:)=0;
	
%centering V on Q-grid through difference in x

for ir=1:jdm
    if(free==1)
        dVx_q(ir,1)=0;
    else
        dVx_q(ir,1)=0; %v_btrop(ir,1);
    end
    for ic=2:jdm-1
        dVx_q(ir,ic)=(v_btrop(ir,ic)-v_btrop(ir,ic-1))/qscx(ir,ic);
    end
    if(free==1)
        dVx_q(ir,idm)=0;
    else
        dVx_q(ir,idm)=0;%v_btrop(ir,idm-1);
    end
end

%remove Nan on land and allow below interpolation
 u_btrop(:,  1)=0;
 u_btrop(:,idm)=0;
%centering U on Q-grid through difference in y

for ic=1:idm
    if(free==1)
        dUy_q(1,ic)=0;
    else
        dUy_q(1,ic)=0; %u_btrop(1,ic);
    end
    for ir=2:idm-1
	dUy_q(ir,ic)=(u_btrop(ir,ic)-u_btrop(ir-1,ic))/qscy(ir,ic);
    end
    if(free==1)
        dUy_q(jdm,1)=0;
    else
        dUy_q(jdm,1)=0; %u_btrop(jdm-1,ic);
    end
end

vort=dVx_q-dUy_q;




%elliptic solver

maxiter=100000;

tollerance=1e-4;

%initialization
%note that because iteration is in the interior this sets boundary conditions

PSIold=zeros(jdm,idm);
PSInew=PSIold;

invrad=pi/180;
R=6.37e6;      %average Earth radius

if(spheric==0)
 for iter=1:maxiter
  rms=0;
%simplest version with dx different from dy
  for ir=2:jdm-1
    for ic=2:idm-1
        b=   qscx(ir,ic)*qscx(ir,ic);
        a=b/(qscy(ir,ic)*qscy(ir,ic));
        PSInew(ir,ic)=(   PSIold(ir,  ic+1)+PSIold(ir,  ic-1) +...
                       a*(PSIold(ir+1,ic  )+PSIold(ir-1,ic  ))-...
                          b*vort(ir  ,ic  )  )/(  2*(a+1)  );
    end %ic
  end %ir
 rms=sqrt(sum(sum((PSInew-PSIold).^2)));
 PSIold=PSInew;

 if(rms<=tollerance)
    break
 end %rms
 disp(iter); disp(rms);
 end %maxiter
else
 for iter=1:maxiter
%simplest version taking into account spherical laplacian
  for ir=2:jdm-1
   for ic=2:idm-1
          Dlambda =(vlon(ir,  ic)-vlon(ir,ic-1))*invrad;
          Dphi    =(ulat(ir  ,ic)-ulat(ir-1,ic))*invrad;
          Dlambda2=Dlambda*Dlambda;
          Dphi2   =Dphi*Dphi;
          phi=qlat(ir,ic)*invrad;
          a=0.5*sin(phi)*cos(phi)*Dlambda2/Dphi;
          b=(cos(phi)*Dlambda/Dphi)^2;
	  c=(R*cos(phi)*Dlambda)^2;
          PSInew(ir,ic)=(     PSIold(ir,  ic+1)+      PSIold(ir,  ic-1) +...
                        (b+a)*PSIold(ir+1,ic  )+(b-a)*PSIold(ir-1,ic  ) -...
                              c*vort(ir  ,ic  )  )/(  2*(b+1)  );
     
   end %ic
  end %ir
 rms=sqrt(sum(sum((PSInew-PSIold).^2)));
 PSIold=PSInew;

 if(rms<=tollerance)
    break
 end %rms
 disp(iter); disp(rms);
 end %maxiter
end %if spheric


%calculation of the volume streamfunction in sverdrup
PSInew=PSInew.*(depthq+sshq)/1e6;
figure(1)
 clev=[-1:0.1:1];
 [c,h]=contour(qlon,qlat,PSInew,clev); h=clabel(c,h,clev);
 set(h,'BackgroundColor',[1 1 0.6]);
 %caxis([clev(1) clev(end)]);
 %colorbar; 
 xlabel('Longitude');
 ylabel('Latitude');
 if(exist('tstep_str')==1);
    time_str=num2str(timestep);
    title(['Barotropic streamfunction using V (Sv), Expt ',expstr,' Timestep ',time_str]) 
 else    
    title(['Barotropic streamfunction using V (Sv), Expt ',expstr,' Year ',num2str(year),' Day ',num2str(day),' Hour ',num2str(hour)])
 end %if
